==> atlassian_connect_models-0.0.1.tar.gz: lib/src/descriptor/descriptor_model.dart <==
library ac_dart.descriptor_model;

import 'dart:core' hide Pattern;

import 'package:constrain/constrain.dart';
import 'package:option/option.dart';
import 'package:quiver/iterables.dart';
import 'package:stuff/stuff.dart';
import 'package:uri/uri.dart';

abstract class ModuleLocation {
  final String key;

  ModuleLocation(this.key);

  String toString() => key;
}

abstract class Scope {
  final String key;

  Scope(this.key);

  String toString() => key;
}

class CoreScope extends Scope {
  CoreScope._(String key) : super(key);
}

abstract class CoreScopes<S extends Scope> {
  final S read;
  final S write;
  final S delete;
  final S admin;

  CoreScopes(S create(String name))
      : read = create('read'),
        write = create('write'),
        delete = create('delete'),
        admin = create('admin');
}

abstract class WebhookEventIdentifier {
  final String key;

  WebhookEventIdentifier(this.key);

  String toString() => key;
}

class I18nProperty implements Jsonable {
  @NotNull()
  final String value;

  final String i18n;

  I18nProperty({this.value, this.i18n});
  I18nProperty.fromJson(Map json)
      : value = json['value'],
        i18n = json['i18n'];

  Map toJson() => (buildJson..add('value', value)..add('i18n', i18n)).json;
}

abstract class Module implements Jsonable {}

abstract class ModuleList implements Jsonable {
  Iterable<ClassifiedModule> get modules;
}

class CoreModuleTypes {
  static const String webItem = 'webItem';
  static const String configurePage = 'configurePage';
  static const String webHook = 'webHook';
  static const String oauthConsumer = 'oauthConsumer';
  static const String adminPage = 'adminPage';
  static const String generalPage = 'generalPage';
}

class ClassifiedModule {
  final String type;
  final Module module;

  ClassifiedModule(this.type, this.module);
}

abstract class BaseModuleList implements ModuleList {
  final Iterable<ClassifiedModule> modules;

  Iterable<Module> filtered(String type) =>
      modules.where((cm) => cm.type == type).map((cm) => cm.module);
  Option<Module> single(String type) =>
      new Option(filtered(type).firstWhere((_) => true, orElse: () => null));

  static Iterable<ClassifiedModule> mapModules(
          Iterable<Module> modules, String type) =>
      modules.map((m) => new ClassifiedModule(type, m));

  static Iterable<ClassifiedModule> mapSingle(Module module, String type) =>
      module != null ? [new ClassifiedModule(type, module)] : const [];

  BaseModuleList(this.modules);
}

class CoreModuleList extends BaseModuleList {
  Iterable<WebItem> get webItems => filtered(CoreModuleTypes.webItem);
  Iterable<Page> get generalPages => filtered(CoreModuleTypes.generalPage);
  Iterable<Page> get adminPages => filtered(CoreModuleTypes.adminPage);
  Option<Page> get configurePage => single(CoreModuleTypes.configurePage);
  Iterable<Webhook> get webhooks => filtered(CoreModuleTypes.webHook);
  Option<OAuthConsumer> get oauthConsumer =>
      single(CoreModuleTypes.oauthConsumer);

  static Iterable<ClassifiedModule> createCoreModules(
      Iterable<Page> generalPages,
      Iterable<Page> adminPages,
      Iterable<WebItem> webItems,
      Page configurePage,
      Iterable<Webhook> webhooks,
      OAuthConsumer oauthConsumer) {
    return concat([
      BaseModuleList.mapModules(generalPages, CoreModuleTypes.generalPage),
      BaseModuleList.mapModules(adminPages, CoreModuleTypes.adminPage),
      BaseModuleList.mapModules(webItems, CoreModuleTypes.webItem),
      BaseModuleList.mapModules(webhooks, CoreModuleTypes.webHook),
      BaseModuleList.mapSingle(configurePage, CoreModuleTypes.configurePage),
      BaseModuleList.mapSingle(oauthConsumer, CoreModuleTypes.oauthConsumer)
    ]);
  }

  CoreModuleList.fromModules(Iterable<ClassifiedModule> modules)
      : super(modules);

  CoreModuleList(
      {Iterable<Page> generalPages: const [],
      Iterable<Page> adminPages: const [],
      Iterable<WebItem> webItems: const [],
      Page configurePage,
      Iterable<Webhook> webhooks: const [],
      OAuthConsumer oauthConsumer})
      : this.fromModules(createCoreModules(generalPages, adminPages, webItems,
            configurePage, webhooks, oauthConsumer));

  CoreModuleList.fromJson(Map json)
      : this(
            generalPages: _fromJsonList(
                json['generalPages'], (wi) => new Page.fromJson(wi)),
            adminPages: _fromJsonList(
                json['adminPages'], (wi) => new Page.fromJson(wi)),
            webItems: _fromJsonList(
                json['webItems'], (wi) => new WebItem.fromJson(wi)),
            webhooks: _fromJsonList(
                json['webhooks'], (wi) => new Webhook.fromJson(wi)),
            configurePage: new Page.fromJson(json['configurePage']),
            oauthConsumer: new OAuthConsumer.fromJson(json['oauthConsumer']));

  Map toJson() => (buildJson
        ..add('generalPages', generalPages)
        ..add('adminPages', adminPages)
        ..add('webItems', webItems)
        ..add('webhooks', webhooks)
        ..add('configurePage', configurePage)
        ..add('oauthConsumer', oauthConsumer))
      .json;
}

abstract class NamedModule implements Module {
  @NotNull()
  final I18nProperty name;

  NamedModule(this.name);
  NamedModule.fromJson(Map json)
      : name = new I18nProperty.fromJson(json['name']);

  Map toJson() => {'name': name.toJson()};
}

abstract class KeyedModule extends NamedModule {
  @NotNull()
  @Pattern(_keyRegexp, isRegExp: true)
  final String key;

  KeyedModule(this.key, I18nProperty name) : super(name);
  KeyedModule.fromJson(Map json)
      : key = json['key'],
        super.fromJson(json);

  Map toJson() => super.toJson()..addAll({'key': key});
}

abstract class Condition implements Jsonable {
  factory Condition.fromJson(Map json) {}

  Map toJson() => {};
}

abstract class ConditionalModule extends KeyedModule {
  @NotNull()
  final List<Condition> conditions;

  ConditionalModule(String key, I18nProperty name, this.conditions)
      : super(key, name);

  ConditionalModule.fromJson(Map json)
      : conditions =
            _fromJsonList(json['conditions'], (j) => new Condition.fromJson(j)),
        super.fromJson(json);

  Map toJson() => (buildJson
        ..addAll(super.toJson())
        ..add('conditions', conditions))
      .json;
}

abstract class BasicWebLink extends ConditionalModule {
  @NotNull()
  final UriTemplate url;

  // TODO: should this be a ModuleLocation??
  final String location;

  @NotNull()
  final int weight;

  BasicWebLink(String key, I18nProperty name, List<Condition> conditions,
      UriTemplate url, String urlStr, this.location, this.weight)
      : this.url = _uriTemplate(url, urlStr),
        super(key, name, conditions);

  BasicWebLink.fromJson(Map json)
      : url = new UriTemplate(json['url']),
        location = json['location'],
        weight = json['weight'],
        super.fromJson(json);

  Map toJson() => (buildJson
        ..addAll(super.toJson())
        ..add('url', url)
        ..add('location', location)
        ..add('weight', weight))
      .json;

  BasicWebLink change({UriTemplate url});
}

class WebItem extends BasicWebLink {
  @NotNull()
  String get location => super.location;

  WebItem(
      {String key,
      I18nProperty name,
      List<Condition> conditions: const [],
      UriTemplate url,
      String urlStr,
      String location,
      int weight: 100})
      : super(key, name, conditions, url, urlStr, location, weight);

  WebItem.fromJson(Map json) : super.fromJson(json);

  WebItem change({UriTemplate url}) => new WebItem(
      key: key,
      name: name,
      conditions: conditions,
      url: url != null ? url : this.url,
      location: location,
      weight: weight);
}

class Webhook implements Module {
  final String event;
  final Uri url;
  final Map<String, Object> params;

  Webhook({this.event, this.url, this.params});

  factory Webhook.fromJson(Map json) {
    final p = parseJson(json);
    return new Webhook(
        event: p.single('event'),
        url: p.single('url', (v) => Uri.parse(v)),
        params: p.single('params'));
  }

  Map toJson() =>
      (buildJson..add('event', event)..add('url', url)..add('params', params))
          .json;
}

class Vendor implements Jsonable {
  final String name;
  final Uri url;

  Vendor({this.name, this.url});

  factory Vendor.fromJson(Map json) {
    final p = parseJson(json);
    return new Vendor(
        name: p.single('name'), url: p.single('url', (v) => Uri.parse(v)));
  }

  Map toJson() => (buildJson..add('name', name)..add('url', url)).json;
}

class Page extends BasicWebLink {
  Page(
      {String key,
      I18nProperty name,
      List<Condition> conditions: const [],
      UriTemplate url,
      String urlStr,
      String location,
      int weight: 100})
      : super(key, name, conditions, url, urlStr, location, weight);

  Page change({UriTemplate url}) => new Page(
      key: key,
      name: name,
      conditions: conditions,
      url: url != null ? url : this.url,
      location: location,
      weight: weight);

  Page.fromJson(Map json) : super.fromJson(json);
}

class AuthenticationType {
  final String _literal;

  const AuthenticationType._internal(this._literal);

  static const AuthenticationType NONE =
      const AuthenticationType._internal('NONE');
  static const AuthenticationType JWT =
      const AuthenticationType._internal('JWT');

  static final Map<String, AuthenticationType> _literalMap = {
    NONE._literal: NONE,
    JWT._literal: JWT
  };

  String toString() => _literal;

  static Option<AuthenticationType> lookup(String literal) =>
      new Option(_literalMap[literal]);
}

class Authentication implements Jsonable {
  @NotNull()
  final AuthenticationType type;

  Authentication({this.type: AuthenticationType.JWT});

  Authentication.fromJson(Map json)
      : type = AuthenticationType
            .lookup(json['type'])
            .getOrElse(() => AuthenticationType.JWT);

  Map toJson() => (buildJson..add('type', type)).json;
}

class Lifecycle implements Jsonable {
  final Uri installed;

  Lifecycle({Uri installed, String installedStr})
      : this.installed = _uri(installed, installedStr);

  Lifecycle.fromJson(Map json) : installed = _uri(null, json['installed']);

  Map toJson() => (buildJson..add('installed', installed)).json;
}

class OAuthConsumer implements Module {
  final String clientId;

  OAuthConsumer(this.clientId);

  OAuthConsumer.fromJson(Map json) : this(json['clientId']);

  Map toJson() => (buildJson..add('clientId', clientId)).json;
}

abstract class BaseConnectAddon<M extends ModuleList> implements Jsonable {
  @NotNull()
  @Pattern(_keyRegexp, isRegExp: true)
  final String key;

  @NotNull()
  final String name;

  @NotNull()
  final String description;

  @NotNull()
  M get modules;

  BaseConnectAddon({this.key, this.name, this.description});

  BaseConnectAddon._fromJson(JsonParser p)
      : key = p.single('key'),
        name = p.single('name'),
        description = p.single('description');

  BaseConnectAddon.fromJson(Map json) : this._fromJson(parseJson(json));
}

abstract class ConnectAddon<M extends ModuleList, S extends Scope>
    extends BaseConnectAddon<M> {
  @NotNull()
  final Uri baseUrl;

//  @NotNull()
  final M modules;

  @NotNull()
  final Authentication authentication;

  final Lifecycle lifecycle;

  final Vendor vendor;

  @NotNull()
  final Iterable<S> scopes;

  ConnectAddon(
      {String key,
      String name,
      String description,
      this.modules,
      Uri baseUrl,
      String baseUrlStr,
      this.authentication,
      this.lifecycle,
      this.vendor,
      this.scopes})
      : this.baseUrl = _uri(baseUrl, baseUrlStr),
        super(key: key, name: name, description: description);

  ConnectAddon._fromJson(
      JsonParser p, M modulesFromJson(Map j), S scopeFromJson(String key))
      : baseUrl = p.single('baseUrl', (s) => Uri.parse(s)),
        modules = p.single('modules', modulesFromJson),
        authentication =
            p.single('authentication', (v) => new Authentication.fromJson(v)),
        vendor = p.single('vendor', (v) => new Vendor.fromJson(v)),
        lifecycle = p.single('lifecycle', (v) => new Lifecycle.fromJson(v)),
        scopes = p.list('scopes', scopeFromJson),
        super._fromJson(p);

  ConnectAddon.fromJson(
      Map json, M modulesFromJson(Map j), S scopeFromJson(String key))
      : this._fromJson(parseJson(json), modulesFromJson, scopeFromJson);

  Map toJson() => (buildJson
        ..add('key', key)
        ..add('name', name)
        ..add('description', description)
        ..add('baseUrl', baseUrl)
        ..add('modules', modules)
        ..add('authentication', authentication)
        ..add('lifecycle', lifecycle)
        ..add('vendor', vendor)
        ..add('scopes', scopes))
      .json;
}

class _ProductlessScope extends Scope {
  _ProductlessScope(String key) : super(key);
}

class ProductlessConnectAddon extends ConnectAddon<CoreModuleList, Scope> {
  ProductlessConnectAddon(
      {String key,
      String name,
      String description,
      CoreModuleList modules,
      Uri baseUrl,
      String baseUrlStr,
      Authentication authentication,
      Lifecycle lifecycle})
      : super(
            key: key,
            name: name,
            description: description,
            modules: modules,
            baseUrl: baseUrl,
            baseUrlStr: baseUrlStr,
            authentication: authentication,
            lifecycle: lifecycle);

  ProductlessConnectAddon.fromJson(Map json)
      : super.fromJson(json, (Map j) => new CoreModuleList.fromJson(j),
            (String s) => new _ProductlessScope(s));

  Map toJson() => super.toJson();
}

_fromJsonList(List jl, mapping(e)) =>
    jl != null ? jl.map(mapping).toList(growable: false) : [];

UriTemplate _uriTemplate(UriTemplate uri, String uriStr) =>
    uri != null ? uri : uriStr != null ? new UriTemplate(uriStr) : null;

Uri _uri(Uri uri, String uriStr) =>
    uri != null ? uri : uriStr != null ? Uri.parse(uriStr) : null;

Map _cleanJson(Map j) {
  final cleaned = {};
  j.forEach((k, v) {
    if (v != null) {
      if (v is Map) {
        cleaned[k] = _cleanJson(v);
      } else if (v is Iterable) {
        final l = v.expand(_cleanJsonIterable).toList(growable: false);
        if (l.isNotEmpty) {
          cleaned[k] = l;
        }
      } else {
        cleaned[k] = v;
      }
    }
  });
  return cleaned;
}

Iterable _cleanJsonIterable(j) {
  if (j == null) {
    return [];
  }

  final result = j is Map ? _cleanJson(j) : j;
  return [result];
}

const String _keyRegexp = r'^[a-zA-Z0-9-\.]+$';

==> atlassian_connect_models-0.0.2.tar.gz: lib/src/descriptor/descriptor_model.dart <==
library ac_dart.descriptor_model;

import 'dart:core' hide Pattern;

import 'package:constrain/constrain.dart';
import 'package:option/option.dart';
import 'package:quiver/iterables.dart';
import 'package:stuff/stuff.dart';
import 'package:uri/uri.dart';

abstract class ModuleLocation {
  final String key;

  ModuleLocation(this.key);

  String toString() => key;
}

abstract class Scope {
  final String key;

  Scope(this.key);

  String toString() => key;
}

class CoreScope extends Scope {
  CoreScope._(String key) : super(key);
}

abstract class CoreScopes<S extends Scope> {
  final S read;
  final S write;
  final S delete;
  final S admin;

  CoreScopes(S create(String name))
      : read = create('read'),
        write = create('write'),
        delete = create('delete'),
        admin = create('admin');
}

abstract class WebhookEventIdentifier {
  final String key;

  WebhookEventIdentifier(this.key);

  String toString() => key;
}

class I18nProperty implements Jsonable {
  @NotNull()
  final String value;

  final String i18n;

  I18nProperty({this.value, this.i18n});
  I18nProperty.fromJson(Map json)
      : value = json['value'],
        i18n = json['i18n'];

  Map toJson() => (buildJson..add('value', value)..add('i18n', i18n)).json;
}

abstract class Module implements Jsonable {}

abstract class ModuleList implements Jsonable {
  Iterable<ClassifiedModule> get modules;
}

class CoreModuleTypes {
  static const String webItem = 'webItem';
  static const String webPanel = 'webPanel';
  static const String configurePage = 'configurePage';
  static const String webHook = 'webHook';
  static const String oauthConsumer = 'oauthConsumer';
  static const String adminPage = 'adminPage';
  static const String generalPage = 'generalPage';
}

class ClassifiedModule {
  final String type;
  final Module module;

  ClassifiedModule(this.type, this.module);
}

abstract class BaseModuleList implements ModuleList {
  final Iterable<ClassifiedModule> modules;

  Iterable<Module> filtered(String type) =>
      modules.where((cm) => cm.type == type).map((cm) => cm.module);
  Option<Module> single(String type) =>
      new Option(filtered(type).firstWhere((_) => true, orElse: () => null));

  static Iterable<ClassifiedModule> mapModules(
          Iterable<Module> modules, String type) =>
      modules.map((m) => new ClassifiedModule(type, m));

  static Iterable<ClassifiedModule> mapSingle(Module module, String type) =>
      module != null ? [new ClassifiedModule(type, module)] : const [];

  BaseModuleList(this.modules);
}

class CoreModuleList extends BaseModuleList {
  Iterable<WebItem> get webItems => filtered(CoreModuleTypes.webItem);
  Iterable<WebPanel> get webPanels => filtered(CoreModuleTypes.webPanel);
  Iterable<Page> get generalPages => filtered(CoreModuleTypes.generalPage);
  Iterable<Page> get adminPages => filtered(CoreModuleTypes.adminPage);
  Option<Page> get configurePage => single(CoreModuleTypes.configurePage);
  Iterable<Webhook> get webhooks => filtered(CoreModuleTypes.webHook);
  Option<OAuthConsumer> get oauthConsumer =>
      single(CoreModuleTypes.oauthConsumer);

  static Iterable<ClassifiedModule> createCoreModules(
      Iterable<Page> generalPages,
      Iterable<Page> adminPages,
      Iterable<WebItem> webItems,
      Iterable<WebPanel> webPanels,
      Page configurePage,
      Iterable<Webhook> webhooks,
      OAuthConsumer oauthConsumer) {
    return concat([
      BaseModuleList.mapModules(generalPages, CoreModuleTypes.generalPage),
      BaseModuleList.mapModules(adminPages, CoreModuleTypes.adminPage),
      BaseModuleList.mapModules(webItems, CoreModuleTypes.webItem),
      BaseModuleList.mapModules(webPanels, CoreModuleTypes.webPanel),
      BaseModuleList.mapModules(webhooks, CoreModuleTypes.webHook),
      BaseModuleList.mapSingle(configurePage, CoreModuleTypes.configurePage),
      BaseModuleList.mapSingle(oauthConsumer, CoreModuleTypes.oauthConsumer)
    ]);
  }

  CoreModuleList.fromModules(Iterable<ClassifiedModule> modules)
      : super(modules);

  CoreModuleList(
      {Iterable<Page> generalPages: const [],
      Iterable<Page> adminPages: const [],
      Iterable<WebItem> webItems: const [],
      Iterable<WebPanel> webPanels: const [],
      Page configurePage,
      Iterable<Webhook> webhooks: const [],
      OAuthConsumer oauthConsumer})
      : this.fromModules(createCoreModules(generalPages, adminPages, webItems,
            webPanels, configurePage, webhooks, oauthConsumer));

  CoreModuleList.fromJson(Map json)
      : this(
            generalPages: _fromJsonList(
                json['generalPages'], (wi) => new Page.fromJson(wi)),
            adminPages: _fromJsonList(
                json['adminPages'], (wi) => new Page.fromJson(wi)),
            webItems: _fromJsonList(
                json['webItems'], (wi) => new WebItem.fromJson(wi)),
            webhooks: _fromJsonList(
                json['webhooks'], (wi) => new Webhook.fromJson(wi)),
            configurePage: new Page.fromJson(json['configurePage']),
            oauthConsumer: new OAuthConsumer.fromJson(json['oauthConsumer']));

  Map toJson() => (buildJson
        ..add('generalPages', generalPages)
        ..add('adminPages', adminPages)
        ..add('webItems', webItems)
        ..add('webPanels', webPanels)
        ..add('webhooks', webhooks)
        ..add('configurePage', configurePage)
        ..add('oauthConsumer', oauthConsumer))
      .json;
}

abstract class NamedModule implements Module {
  @NotNull()
  final I18nProperty name;

  NamedModule(this.name);
  NamedModule.fromJson(Map json)
      : name = new I18nProperty.fromJson(json['name']);

  Map toJson() => {'name': name.toJson()};
}

abstract class KeyedModule extends NamedModule {
  @NotNull()
  @Pattern(_keyRegexp, isRegExp: true)
  final String key;

  KeyedModule(this.key, I18nProperty name) : super(name);
  KeyedModule.fromJson(Map json)
      : key = json['key'],
        super.fromJson(json);

  Map toJson() => super.toJson()..addAll({'key': key});
}

abstract class Condition implements Jsonable {
  factory Condition.fromJson(Map json) {}

  Map toJson() => {};
}

abstract class ConditionalModule extends KeyedModule {
  @NotNull()
  final List<Condition> conditions;

  ConditionalModule(String key, I18nProperty name, this.conditions)
      : super(key, name);

  ConditionalModule.fromJson(Map json)
      : conditions =
            _fromJsonList(json['conditions'], (j) => new Condition.fromJson(j)),
        super.fromJson(json);

  Map toJson() => (buildJson
        ..addAll(super.toJson())
        ..add('conditions', conditions))
      .json;
}

abstract class BasicWebLink extends ConditionalModule {
  @NotNull()
  final UriTemplate url;

  // TODO: should this be a ModuleLocation??
  final String location;

  @NotNull()
  final int weight;

  BasicWebLink(String key, I18nProperty name, List<Condition> conditions,
      UriTemplate url, String urlStr, this.location, this.weight)
      : this.url = _uriTemplate(url, urlStr),
        super(key, name, conditions);

  BasicWebLink.fromJson(Map json)
      : url = new UriTemplate(json['url']),
        location = json['location'],
        weight = json['weight'],
        super.fromJson(json);

  Map toJson() => (buildJson
        ..addAll(super.toJson())
        ..add('url', url)
        ..add('location', location)
        ..add('weight', weight))
      .json;

  BasicWebLink change({UriTemplate url});
}

class WebItem extends BasicWebLink {
  @NotNull()
  String get location => super.location;

  WebItem(
      {String key,
      I18nProperty name,
      List<Condition> conditions: const [],
      UriTemplate url,
      String urlStr,
      String location,
      int weight: 100})
      : super(key, name, conditions, url, urlStr, location, weight);

  WebItem.fromJson(Map json) : super.fromJson(json);

  WebItem change({UriTemplate url}) => new WebItem(
      key: key,
      name: name,
      conditions: conditions,
      url: url != null ? url : this.url,
      location: location,
      weight: weight);
}

class WebPanel extends BasicWebLink {
  @NotNull()
  String get location => super.location;

  WebPanel(
      {String key,
      I18nProperty name,
      List<Condition> conditions: const [],
      UriTemplate url,
      String urlStr,
      String location,
      int weight: 100})
      : super(key, name, conditions, url, urlStr, location, weight);

  WebPanel.fromJson(Map json) : super.fromJson(json);

  WebPanel change({UriTemplate url}) => new WebPanel(
      key: key,
      name: name,
      conditions: conditions,
      url: url != null ? url : this.url,
      location: location,
      weight: weight);
}

class Webhook implements Module {
  final String event;
  final Uri url;
  final Map<String, Object> params;

  Webhook({this.event, this.url, this.params});

  factory Webhook.fromJson(Map json) {
    final p = parseJson(json);
    return new Webhook(
        event: p.single('event'),
        url: p.single('url', (v) => Uri.parse(v)),
        params: p.single('params'));
  }

  Map toJson() =>
      (buildJson..add('event', event)..add('url', url)..add('params', params))
          .json;
}

class Vendor implements Jsonable {
  final String name;
  final Uri url;

  Vendor({this.name, this.url});

  factory Vendor.fromJson(Map json) {
    final p = parseJson(json);
    return new Vendor(
        name: p.single('name'), url: p.single('url', (v) => Uri.parse(v)));
  }

  Map toJson() => (buildJson..add('name', name)..add('url', url)).json;
}

class Page extends BasicWebLink {
  Page(
      {String key,
      I18nProperty name,
      List<Condition> conditions: const [],
      UriTemplate url,
      String urlStr,
      String location,
      int weight: 100})
      : super(key, name, conditions, url, urlStr, location, weight);

  Page change({UriTemplate url}) => new Page(
      key: key,
      name: name,
      conditions: conditions,
      url: url != null ? url : this.url,
      location: location,
      weight: weight);

  Page.fromJson(Map json) : super.fromJson(json);
}

class AuthenticationType {
  final String _literal;

  const AuthenticationType._internal(this._literal);

  static const AuthenticationType NONE =
      const AuthenticationType._internal('none');
  static const AuthenticationType JWT =
      const AuthenticationType._internal('jwt');

  static final Map<String, AuthenticationType> _literalMap = {
    NONE._literal: NONE,
    JWT._literal: JWT
  };

  String toString() => _literal;

  static Option<AuthenticationType> lookup(String literal) =>
      new Option(_literalMap[literal.toLowerCase()]);
}

class Authentication implements Jsonable {
  @NotNull()
  final AuthenticationType type;

  Authentication({this.type: AuthenticationType.JWT});

  Authentication.fromJson(Map json)
      : type = AuthenticationType
            .lookup(json['type'])
            .getOrElse(() => AuthenticationType.JWT);

  Map toJson() => (buildJson..add('type', type)).json;
}

class Lifecycle implements Jsonable {
  final Uri installed;

  Lifecycle({Uri installed, String installedStr})
      : this.installed = _uri(installed, installedStr);

  Lifecycle.fromJson(Map json) : installed = _uri(null, json['installed']);

  Map toJson() => (buildJson..add('installed', installed)).json;
}

class OAuthConsumer implements Module {
  final String clientId;

  OAuthConsumer(this.clientId);

  OAuthConsumer.fromJson(Map json) : this(json['clientId']);

  Map toJson() => (buildJson..add('clientId', clientId)).json;
}

abstract class BaseConnectAddon<M extends ModuleList> implements Jsonable {
  @NotNull()
  @Pattern(_keyRegexp, isRegExp: true)
  final String key;

  @NotNull()
  final String name;

  @NotNull()
  final String description;

  @NotNull()
  M get modules;

  @NotNull()
  final bool enableLicensing;

  BaseConnectAddon(
      {this.key, this.name, this.description, this.enableLicensing});

  BaseConnectAddon._fromJson(JsonParser p)
      : key = p.single('key'),
        name = p.single('name'),
        description = p.single('description'),
        enableLicensing = p.single('enableLicensing');

  BaseConnectAddon.fromJson(Map json) : this._fromJson(parseJson(json));
}

abstract class ConnectAddon<M extends ModuleList, S extends Scope>
    extends BaseConnectAddon<M> {
  @NotNull()
  final Uri baseUrl;

//  @NotNull()
  final M modules;

  @NotNull()
  final Authentication authentication;

  final Lifecycle lifecycle;

  final Vendor vendor;

  @NotNull()
  final Iterable<S> scopes;

  ConnectAddon(
      {String key,
      String name,
      String description,
      this.modules,
      Uri baseUrl,
      String baseUrlStr,
      this.authentication,
      this.lifecycle,
      this.vendor,
      this.scopes})
      : this.baseUrl = _uri(baseUrl, baseUrlStr),
        super(key: key, name: name, description: description);

  ConnectAddon._fromJson(
      JsonParser p, M modulesFromJson(Map j), S scopeFromJson(String key))
      : baseUrl = p.single('baseUrl', (s) => Uri.parse(s)),
        modules = p.single('modules', modulesFromJson),
        authentication =
            p.single('authentication', (v) => new Authentication.fromJson(v)),
        vendor = p.single('vendor', (v) => new Vendor.fromJson(v)),
        lifecycle = p.single('lifecycle', (v) => new Lifecycle.fromJson(v)),
        scopes = p.list('scopes', scopeFromJson),
        super._fromJson(p);

  ConnectAddon.fromJson(
      Map json, M modulesFromJson(Map j), S scopeFromJson(String key))
      : this._fromJson(parseJson(json), modulesFromJson, scopeFromJson);

  Map toJson() => (buildJson
        ..add('key', key)
        ..add('name', name)
        ..add('description', description)
        ..add('enableLicensing', enableLicensing)
        ..add('baseUrl', baseUrl)
        ..add('modules', modules)
        ..add('authentication', authentication)
        ..add('lifecycle', lifecycle)
        ..add('vendor', vendor)
        ..add('scopes', scopes))
      .json;
}

class _ProductlessScope extends Scope {
  _ProductlessScope(String key) : super(key);
}

class ProductlessConnectAddon extends ConnectAddon<CoreModuleList, Scope> {
  ProductlessConnectAddon(
      {String key,
      String name,
      String description,
      CoreModuleList modules,
      Uri baseUrl,
      String baseUrlStr,
      Authentication authentication,
      Lifecycle lifecycle})
      : super(
            key: key,
            name: name,
            description: description,
            modules: modules,
            baseUrl: baseUrl,
            baseUrlStr: baseUrlStr,
            authentication: authentication,
            lifecycle: lifecycle);

  ProductlessConnectAddon.fromJson(Map json)
      : super.fromJson(json, (Map j) => new CoreModuleList.fromJson(j),
            (String s) => new _ProductlessScope(s));

  Map toJson() => super.toJson();
}

_fromJsonList(List jl, mapping(e)) =>
    jl != null ? jl.map(mapping).toList(growable: false) : [];

UriTemplate _uriTemplate(UriTemplate uri, String uriStr) =>
    uri != null ? uri : uriStr != null ? new UriTemplate(uriStr) : null;

Uri _uri(Uri uri, String uriStr) =>
    uri != null ? uri : uriStr != null ? Uri.parse(uriStr) : null;

Map _cleanJson(Map j) {
  final cleaned = {};
  j.forEach((k, v) {
    if (v != null) {
      if (v is Map) {
        cleaned[k] = _cleanJson(v);
      } else if (v is Iterable) {
        final l = v.expand(_cleanJsonIterable).toList(growable: false);
        if (l.isNotEmpty) {
          cleaned[k] = l;
        }
      } else {
        cleaned[k] = v;
      }
    }
  });
  return cleaned;
}

Iterable _cleanJsonIterable(j) {
  if (j == null) {
    return [];
  }

  final result = j is Map ? _cleanJson(j) : j;
  return [result];
}

const String _keyRegexp = r'^[a-zA-Z0-9-\.]+$';
